import asyncio
from typing import Generic, TypeVar

import pytest

from tests.util import requires_grpclib, requires_protobuf  # noqa: F401


@pytest.mark.asyncio
async def test_grpclib_reflection(requires_grpclib, requires_protobuf):
    from google.protobuf import descriptor_pb2
    from grpclib.reflection.service import ServerReflection
    from grpclib.reflection.v1.reflection_grpc import ServerReflectionBase as ServerReflectionBaseV1
    from grpclib.reflection.v1alpha.reflection_grpc import ServerReflectionBase as ServerReflectionBaseV1Alpha
    from grpclib.testing import ChannelFor

    from tests.outputs.grpclib_reflection.example_service import TestBase
    from tests.outputs.grpclib_reflection.grpc.reflection.v1 import (
        ErrorResponse,
        ListServiceResponse,
        ServerReflectionRequest,
        ServerReflectionStub,
        ServiceResponse,
    )
    from tests.outputs.grpclib_reflection_descriptors.google_proto_descriptor_pool import (
        default_google_proto_descriptor_pool,
    )

    class TestService(TestBase):
        pass

    T = TypeVar("T")

    class AsyncIterableQueue(Generic[T]):
        CLOSED_SENTINEL = object()

        def __init__(self):
            self._queue = asyncio.Queue()
            self._done = asyncio.Event()

        def put(self, item: T):
            self._queue.put_nowait(item)

        def close(self):
            self._queue.put_nowait(self.CLOSED_SENTINEL)

        def __aiter__(self):
            return self

        async def __anext__(self) -> T:
            val = await self._queue.get()
            if val is self.CLOSED_SENTINEL:
                raise StopAsyncIteration
            return val

    service = TestService()
    services = ServerReflection.extend([service])
    for service in services:
        # This won't be needed once https://github.com/vmagamedov/grpclib/pull/204 is in.
        if isinstance(service, ServerReflectionBaseV1Alpha | ServerReflectionBaseV1):
            service._pool = default_google_proto_descriptor_pool

    async with ChannelFor(services) as channel:
        requests = AsyncIterableQueue[ServerReflectionRequest]()
        responses = ServerReflectionStub(channel).server_reflection_info(requests)

        # list services
        requests.put(ServerReflectionRequest(list_services=""))
        response = await anext(responses)
        assert response.list_services_response == ListServiceResponse(
            service=[ServiceResponse(name="example_service.Test")]
        )

        # list methods

        # should fail before we've added descriptors to the protobuf pool
        requests.put(ServerReflectionRequest(file_containing_symbol="example_service.Test"))
        response = await anext(responses)
        assert response.error_response == ErrorResponse(error_code=5, error_message="not found")
        assert response.file_descriptor_response is None

        # now it should work
        import tests.outputs.grpclib_reflection_descriptors.example_service as example_service_with_desc

        requests.put(ServerReflectionRequest(file_containing_symbol="example_service.Test"))
        response = await anext(responses)
        expected = descriptor_pb2.FileDescriptorProto.FromString(
            example_service_with_desc.EXAMPLE_SERVICE_PROTO_DESCRIPTOR.serialized_pb
        )
        assert response.error_response is None
        assert response.file_descriptor_response is not None
        assert len(response.file_descriptor_response.file_descriptor_proto) == 1
        actual = descriptor_pb2.FileDescriptorProto.FromString(
            response.file_descriptor_response.file_descriptor_proto[0]
        )
        assert actual == expected

        requests.close()

        await anext(responses, None)



import pytest


def test_check_compiler_version():
    from betterproto2 import __version__, check_compiler_version

    x, y, z = (int(x) for x in __version__.split("."))

    check_compiler_version(__version__)
    check_compiler_version(f"{x}.{y}.{z - 1}")
    check_compiler_version(f"{x}.{y}.{z + 1}")

    with pytest.raises(ImportError):
        check_compiler_version(f"{x}.{y - 1}.{z}")

    with pytest.raises(ImportError):
        check_compiler_version(f"{x}.{y + 1}.{z}")

    with pytest.raises(ImportError):
        check_compiler_version(f"{x + 1}.{y}.{z}")

    with pytest.raises(ImportError):
        check_compiler_version(f"{x - 1}.{y}.{z}")



# from tests.output_betterproto.map import Enum, Test


def test_map():
    # msg = Test(counts={"a": 1, "b": 2})
    # assert msg == Test.parse(bytes(msg))
    # assert msg == Test.from_dict(msg.to_dict())

    pass
    # msg = Test(map_enum={1: Enum.ONE, 2: Enum.TWO})
    # assert msg == Test.parse(bytes(msg))
    # assert msg == Test.from_dict(msg.to_dict())



from tests.util import requires_grpcio, requires_grpclib  # noqa: F401


def test_all_definition(requires_grpclib, requires_grpcio):
    """
    Check that a compiled module defines __all__ with the right value.

    These modules have been chosen since they contain messages, services and enums.
    """
    import tests.outputs.enum.enum as enum
    import tests.outputs.service.service as service

    assert service.__all__ == (
        "DoThingRequest",
        "DoThingResponse",
        "GetThingRequest",
        "GetThingResponse",
        "TestBase",
        "TestStub",
        "TestSyncStub",
        "ThingType",
    )
    assert enum.__all__ == (
        "ArithmeticOperator",
        "Choice",
        "EnumMessage",
        "HttpCode",
        "NewVersion",
        "NewVersionMessage",
        "NoStriping",
        "OldVersion",
        "OldVersionMessage",
        "Test",
    )



import pytest

from tests.util import requires_protobuf  # noqa: F401


def test_message_enum_descriptors(requires_protobuf):
    from tests.outputs.import_cousin_package_same_name.import_cousin_package_same_name.test.subpackage import Test

    # importing the cousin should cause no descriptor pool errors since the subpackage imports it once already
    from tests.outputs.import_cousin_package_same_name_descriptors.import_cousin_package_same_name.cousin.subpackage import (  # noqa: E501
        CousinMessage,
    )
    from tests.outputs.import_cousin_package_same_name_descriptors.import_cousin_package_same_name.test.subpackage import (  # noqa: E501
        Test as TestWithDesc,
    )

    # Normally descriptors are not available as they require protobuf support
    # to inteoperate with other libraries.
    with pytest.raises(AttributeError):
        Test.DESCRIPTOR.full_name

    # But the python_betterproto2_opt=google_protobuf_descriptors option
    # will add them in as long as protobuf is depended on.
    assert TestWithDesc.DESCRIPTOR.full_name == "import_cousin_package_same_name.test.subpackage.Test"
    assert CousinMessage.DESCRIPTOR.full_name == "import_cousin_package_same_name.cousin.subpackage.CousinMessage"



from collections.abc import AsyncIterator

import pytest

from tests.util import requires_grpclib  # noqa: F401


@pytest.mark.asyncio
async def test_calls_with_different_cardinalities(requires_grpclib):
    from grpclib.testing import ChannelFor

    from tests.outputs.example_service.example_service import (
        ExampleRequest,
        ExampleResponse,
        TestBase,
        TestStub,
    )

    class ExampleService(TestBase):
        async def example_unary_unary(self, example_request: ExampleRequest) -> "ExampleResponse":
            return ExampleResponse(
                example_string=example_request.example_string,
                example_integer=example_request.example_integer,
            )

        async def example_unary_stream(self, example_request: ExampleRequest) -> AsyncIterator["ExampleResponse"]:
            response = ExampleResponse(
                example_string=example_request.example_string,
                example_integer=example_request.example_integer,
            )
            yield response
            yield response
            yield response

        async def example_stream_unary(
            self, example_request_iterator: AsyncIterator["ExampleRequest"]
        ) -> "ExampleResponse":
            async for example_request in example_request_iterator:
                return ExampleResponse(
                    example_string=example_request.example_string,
                    example_integer=example_request.example_integer,
                )

        async def example_stream_stream(
            self, example_request_iterator: AsyncIterator["ExampleRequest"]
        ) -> AsyncIterator["ExampleResponse"]:
            async for example_request in example_request_iterator:
                yield ExampleResponse(
                    example_string=example_request.example_string,
                    example_integer=example_request.example_integer,
                )

    example_request = ExampleRequest("test string", 42)

    async with ChannelFor([ExampleService()]) as channel:
        stub = TestStub(channel)

        # unary unary
        response = await stub.example_unary_unary(example_request)
        assert response.example_string == example_request.example_string
        assert response.example_integer == example_request.example_integer

        # unary stream
        async for response in stub.example_unary_stream(example_request):
            assert response.example_string == example_request.example_string
            assert response.example_integer == example_request.example_integer

        # stream unary
        async def request_iterator():
            yield example_request
            yield example_request
            yield example_request

        response = await stub.example_stream_unary(request_iterator())
        assert response.example_string == example_request.example_string
        assert response.example_integer == example_request.example_integer

        # stream stream
        async for response in stub.example_stream_stream(request_iterator()):
            assert response.example_string == example_request.example_string
            assert response.example_integer == example_request.example_integer



from io import BytesIO
from pathlib import Path
from shutil import which
from subprocess import run

import pytest

import betterproto2
from tests.outputs.map import map
from tests.outputs.nested import nested
from tests.outputs.oneof import oneof
from tests.outputs.repeated import repeated
from tests.outputs.repeatedpacked import repeatedpacked

oneof_example = oneof.Test().from_dict({"pitied": 1, "just_a_regular_field": 123456789, "bar_name": "Testing"})

len_oneof = len(bytes(oneof_example))

nested_example = nested.Test().from_dict(
    {
        "nested": {"count": 1},
        "sibling": {"foo": 2},
        "sibling2": {"foo": 3},
        "msg": nested.TestMsg.THIS,
    }
)

repeated_example = repeated.Test().from_dict({"names": ["blah", "Blah2"]})

packed_example = repeatedpacked.Test().from_dict(
    {"counts": [1, 2, 3], "signed": [-1, 2, -3], "fixed": [1.2, -2.3, 3.4]}
)

map_example = map.Test().from_dict({"counts": {"blah": 1, "Blah2": 2}})

streams_path = Path("tests/streams/")

java = which("java")


def test_load_varint_too_long():
    with BytesIO(b"\x80\x80\x80\x80\x80\x80\x80\x80\x80\x80\x01") as stream, pytest.raises(ValueError):
        betterproto2.load_varint(stream)

    with BytesIO(b"\x80\x80\x80\x80\x80\x80\x80\x80\x80\x01") as stream:
        # This should not raise a ValueError, as it is within 64 bits
        betterproto2.load_varint(stream)


def test_load_varint_file():
    with open(streams_path / "message_dump_file_single.expected", "rb") as stream:
        assert betterproto2.load_varint(stream) == (8, b"\x08")  # Single-byte varint
        stream.read(2)  # Skip until first multi-byte
        assert betterproto2.load_varint(stream) == (
            123456789,
            b"\x95\x9a\xef\x3a",
        )  # Multi-byte varint


def test_load_varint_cutoff():
    with open(streams_path / "load_varint_cutoff.in", "rb") as stream:
        with pytest.raises(EOFError):
            betterproto2.load_varint(stream)

        stream.seek(1)
        with pytest.raises(EOFError):
            betterproto2.load_varint(stream)


def test_dump_varint_file(tmp_path):
    # Dump test varints to file
    with open(tmp_path / "dump_varint_file.out", "wb") as stream:
        betterproto2.dump_varint(8, stream)  # Single-byte varint
        betterproto2.dump_varint(123456789, stream)  # Multi-byte varint

    # Check that file contents are as expected
    with (
        open(tmp_path / "dump_varint_file.out", "rb") as test_stream,
        open(streams_path / "message_dump_file_single.expected", "rb") as exp_stream,
    ):
        assert betterproto2.load_varint(test_stream) == betterproto2.load_varint(exp_stream)
        exp_stream.read(2)
        assert betterproto2.load_varint(test_stream) == betterproto2.load_varint(exp_stream)


def test_parse_fields():
    with open(streams_path / "message_dump_file_single.expected", "rb") as stream:
        parsed_bytes = betterproto2.parse_fields(stream.read())

    with open(streams_path / "message_dump_file_single.expected", "rb") as stream:
        parsed_stream = betterproto2.load_fields(stream)
        for field in parsed_bytes:
            assert field == next(parsed_stream)


def test_message_dump_file_single(tmp_path):
    # Write the message to the stream
    with open(tmp_path / "message_dump_file_single.out", "wb") as stream:
        oneof_example.dump(stream)

    # Check that the outputted file is exactly as expected
    with (
        open(tmp_path / "message_dump_file_single.out", "rb") as test_stream,
        open(streams_path / "message_dump_file_single.expected", "rb") as exp_stream,
    ):
        assert test_stream.read() == exp_stream.read()


def test_message_dump_file_multiple(tmp_path):
    # Write the same Message twice and another, different message
    with open(tmp_path / "message_dump_file_multiple.out", "wb") as stream:
        oneof_example.dump(stream)
        oneof_example.dump(stream)
        nested_example.dump(stream)

    # Check that all three Messages were outputted to the file correctly
    with (
        open(tmp_path / "message_dump_file_multiple.out", "rb") as test_stream,
        open(streams_path / "message_dump_file_multiple.expected", "rb") as exp_stream,
    ):
        assert test_stream.read() == exp_stream.read()


def test_message_dump_delimited(tmp_path):
    with open(tmp_path / "message_dump_delimited.out", "wb") as stream:
        oneof_example.dump(stream, True)
        oneof_example.dump(stream, True)
        nested_example.dump(stream, True)

    with (
        open(tmp_path / "message_dump_delimited.out", "rb") as test_stream,
        open(streams_path / "delimited_messages.in", "rb") as exp_stream,
    ):
        assert test_stream.read() == exp_stream.read()


def test_message_load_file_single():
    with open(streams_path / "message_dump_file_single.expected", "rb") as stream:
        assert oneof.Test().load(stream) == oneof_example
        stream.seek(0)
        assert oneof.Test().load(stream, len_oneof) == oneof_example


def test_message_load_file_multiple():
    with open(streams_path / "message_dump_file_multiple.expected", "rb") as stream:
        oneof_size = len_oneof
        assert oneof.Test().load(stream, oneof_size) == oneof_example
        assert oneof.Test().load(stream, oneof_size) == oneof_example
        assert nested.Test().load(stream) == nested_example
        assert stream.read(1) == b""


def test_message_load_too_small():
    with open(streams_path / "message_dump_file_single.expected", "rb") as stream, pytest.raises(ValueError):
        oneof.Test().load(stream, len_oneof - 1)


def test_message_load_delimited():
    with open(streams_path / "delimited_messages.in", "rb") as stream:
        assert oneof.Test().load(stream, betterproto2.SIZE_DELIMITED) == oneof_example
        assert oneof.Test().load(stream, betterproto2.SIZE_DELIMITED) == oneof_example
        assert nested.Test().load(stream, betterproto2.SIZE_DELIMITED) == nested_example
        assert stream.read(1) == b""


def test_message_load_too_large():
    with open(streams_path / "message_dump_file_single.expected", "rb") as stream, pytest.raises(ValueError):
        oneof.Test().load(stream, len_oneof + 1)


def test_calculate_varint_size_negative():
    single_byte = -1
    multi_byte = -10000000
    edge = -(1 << 63)
    before = -(1 << 63) + 1

    assert len(betterproto2.encode_varint(single_byte)) == 10
    assert len(betterproto2.encode_varint(multi_byte)) == 10
    assert len(betterproto2.encode_varint(edge)) == 10
    assert len(betterproto2.encode_varint(before)) == 10


def test_calculate_varint_size_positive():
    single_byte = 1
    multi_byte = 10000000

    assert len(betterproto2.encode_varint(single_byte))
    assert len(betterproto2.encode_varint(multi_byte))


def test_dump_varint_negative(tmp_path):
    single_byte = -1
    multi_byte = -10000000
    edge = -(1 << 63)
    beyond = -(1 << 63) - 1
    before = -(1 << 63) + 1

    with open(tmp_path / "dump_varint_negative.out", "wb") as stream:
        betterproto2.dump_varint(single_byte, stream)
        betterproto2.dump_varint(multi_byte, stream)
        betterproto2.dump_varint(edge, stream)
        betterproto2.dump_varint(before, stream)

        with pytest.raises(ValueError):
            betterproto2.dump_varint(beyond, stream)

    with (
        open(streams_path / "dump_varint_negative.expected", "rb") as exp_stream,
        open(tmp_path / "dump_varint_negative.out", "rb") as test_stream,
    ):
        assert test_stream.read() == exp_stream.read()


def test_dump_varint_positive(tmp_path):
    single_byte = 1
    multi_byte = 10000000

    with open(tmp_path / "dump_varint_positive.out", "wb") as stream:
        betterproto2.dump_varint(single_byte, stream)
        betterproto2.dump_varint(multi_byte, stream)

    with (
        open(tmp_path / "dump_varint_positive.out", "rb") as test_stream,
        open(streams_path / "dump_varint_positive.expected", "rb") as exp_stream,
    ):
        assert test_stream.read() == exp_stream.read()


# Java compatibility tests


@pytest.fixture(scope="module")
def compile_jar():
    # Skip if not all required tools are present
    if java is None:
        pytest.skip("`java` command is absent and is required")
    mvn = which("mvn")
    if mvn is None:
        pytest.skip("Maven is absent and is required")

    # Compile the JAR
    proc_maven = run([mvn, "clean", "install", "-f", "tests/streams/java/pom.xml"])
    if proc_maven.returncode != 0:
        pytest.skip("Maven compatibility-test.jar build failed (maybe Java version <11?)")


jar = "tests/streams/java/target/compatibility-test.jar"


def run_jar(command: str, tmp_path):
    return run([java, "-jar", jar, command, tmp_path], check=True)


def run_java_single_varint(value: int, tmp_path) -> int:
    # Write single varint to file
    with open(tmp_path / "py_single_varint.out", "wb") as stream:
        betterproto2.dump_varint(value, stream)

    # Have Java read this varint and write it back
    run_jar("single_varint", tmp_path)

    # Read single varint from Java output file
    with open(tmp_path / "java_single_varint.out", "rb") as stream:
        returned = betterproto2.load_varint(stream)
        with pytest.raises(EOFError):
            betterproto2.load_varint(stream)

    return returned


def test_single_varint(compile_jar, tmp_path):
    single_byte = (1, b"\x01")
    multi_byte = (123456789, b"\x95\x9a\xef\x3a")

    # Write a single-byte varint to a file and have Java read it back
    returned = run_java_single_varint(single_byte[0], tmp_path)
    assert returned == single_byte

    # Same for a multi-byte varint
    returned = run_java_single_varint(multi_byte[0], tmp_path)
    assert returned == multi_byte


def test_multiple_varints(compile_jar, tmp_path):
    single_byte = (1, b"\x01")
    multi_byte = (123456789, b"\x95\x9a\xef\x3a")
    over32 = (3000000000, b"\x80\xbc\xc1\x96\x0b")

    # Write two varints to the same file
    with open(tmp_path / "py_multiple_varints.out", "wb") as stream:
        betterproto2.dump_varint(single_byte[0], stream)
        betterproto2.dump_varint(multi_byte[0], stream)
        betterproto2.dump_varint(over32[0], stream)

    # Have Java read these varints and write them back
    run_jar("multiple_varints", tmp_path)

    # Read varints from Java output file
    with open(tmp_path / "java_multiple_varints.out", "rb") as stream:
        returned_single = betterproto2.load_varint(stream)
        returned_multi = betterproto2.load_varint(stream)
        returned_over32 = betterproto2.load_varint(stream)
        with pytest.raises(EOFError):
            betterproto2.load_varint(stream)

    assert returned_single == single_byte
    assert returned_multi == multi_byte
    assert returned_over32 == over32


def test_single_message(compile_jar, tmp_path):
    # Write message to file
    with open(tmp_path / "py_single_message.out", "wb") as stream:
        oneof_example.dump(stream)

    # Have Java read and return the message
    run_jar("single_message", tmp_path)

    # Read and check the returned message
    with open(tmp_path / "java_single_message.out", "rb") as stream:
        returned = oneof.Test().load(stream, len(bytes(oneof_example)))
        assert stream.read() == b""

    assert returned == oneof_example


def test_multiple_messages(compile_jar, tmp_path):
    # Write delimited messages to file
    with open(tmp_path / "py_multiple_messages.out", "wb") as stream:
        oneof_example.dump(stream, True)
        nested_example.dump(stream, True)

    # Have Java read and return the messages
    run_jar("multiple_messages", tmp_path)

    # Read and check the returned messages
    with open(tmp_path / "java_multiple_messages.out", "rb") as stream:
        returned_oneof = oneof.Test().load(stream, betterproto2.SIZE_DELIMITED)
        returned_nested = nested.Test().load(stream, betterproto2.SIZE_DELIMITED)
        assert stream.read() == b""

    assert returned_oneof == oneof_example
    assert returned_nested == nested_example


def test_infinite_messages(compile_jar, tmp_path):
    num_messages = 5

    # Write delimited messages to file
    with open(tmp_path / "py_infinite_messages.out", "wb") as stream:
        for _ in range(num_messages):
            oneof_example.dump(stream, True)

    # Have Java read and return the messages
    run_jar("infinite_messages", tmp_path)

    # Read and check the returned messages
    messages = []
    with open(tmp_path / "java_infinite_messages.out", "rb") as stream:
        while True:
            try:
                messages.append(oneof.Test().load(stream, betterproto2.SIZE_DELIMITED))
            except EOFError:
                break

    assert len(messages) == num_messages



from datetime import datetime, timedelta

import pytest

import tests.outputs.googletypes_request.google.protobuf as protobuf
from tests.util import requires_grpclib  # noqa: F401


@pytest.mark.asyncio
async def test_channel_receives_wrapped_type(requires_grpclib):
    from tests.mocks import MockChannel
    from tests.outputs.googletypes_request.googletypes_request import Input, TestStub

    test_cases = [
        (TestStub.send_double, protobuf.DoubleValue, 2.5),
        (TestStub.send_float, protobuf.FloatValue, 2.5),
        (TestStub.send_int_64, protobuf.Int64Value, -64),
        (TestStub.send_u_int_64, protobuf.UInt64Value, 64),
        (TestStub.send_int_32, protobuf.Int32Value, -32),
        (TestStub.send_u_int_32, protobuf.UInt32Value, 32),
        (TestStub.send_bool, protobuf.BoolValue, True),
        (TestStub.send_string, protobuf.StringValue, "string"),
        (TestStub.send_bytes, protobuf.BytesValue, bytes(0xFF)[0:4]),
        (TestStub.send_datetime, protobuf.Timestamp, datetime(2038, 1, 19, 3, 14, 8)),
        (TestStub.send_timedelta, protobuf.Duration, timedelta(seconds=123456)),
    ]

    for service_method, wrapper_class, value in test_cases:
        wrapped_value = wrapper_class()
        wrapped_value.value = value
        channel = MockChannel(responses=[Input()])
        service = TestStub(channel)

        await service_method(service, wrapped_value)

        assert channel.requests[0]["request"] == type(wrapped_value)



import pytest

from tests.util import requires_grpclib  # noqa: F401


@pytest.mark.asyncio
async def test_service_correctly_imports_reference_message(requires_grpclib):
    from tests.mocks import MockChannel
    from tests.outputs.import_service_input_message.import_service_input_message import (
        RequestMessage,
        RequestResponse,
        TestStub,
    )

    mock_response = RequestResponse(value=10)
    service = TestStub(MockChannel([mock_response]))
    response = await service.do_thing(RequestMessage(1))
    assert mock_response == response


@pytest.mark.asyncio
async def test_service_correctly_imports_reference_message_from_child_package(requires_grpclib):
    from tests.mocks import MockChannel
    from tests.outputs.import_service_input_message.import_service_input_message import (
        RequestResponse,
        TestStub,
    )
    from tests.outputs.import_service_input_message.import_service_input_message.child import (
        ChildRequestMessage,
    )

    mock_response = RequestResponse(value=10)
    service = TestStub(MockChannel([mock_response]))
    response = await service.do_thing_2(ChildRequestMessage(1))
    assert mock_response == response


@pytest.mark.asyncio
async def test_service_correctly_imports_nested_reference(requires_grpclib):
    from tests.mocks import MockChannel
    from tests.outputs.import_service_input_message.import_service_input_message import (
        NestedRequestMessage,
        RequestResponse,
        TestStub,
    )

    mock_response = RequestResponse(value=10)
    service = TestStub(MockChannel([mock_response]))
    response = await service.do_thing_3(NestedRequestMessage(1))
    assert mock_response == response



def test_struct_to_dict():
    from tests.outputs.google.google.protobuf import Struct

    struct = Struct.from_dict(
        {
            "null_field": None,
            "number_field": 12,
            "string_field": "test",
            "bool_field": True,
            "struct_field": {"x": "abc"},
            "list_field": [42, False, None],
        }
    )

    assert struct.to_dict() == {
        "null_field": None,
        "number_field": 12,
        "string_field": "test",
        "bool_field": True,
        "struct_field": {"x": "abc"},
        "list_field": [42, False, None],
    }

    assert Struct.from_dict(struct.to_dict()) == struct


def test_listvalue_to_dict():
    from tests.outputs.google.google.protobuf import ListValue

    list_value = ListValue.from_dict([42, False, {}])

    assert list_value.to_dict() == [42, False, {}]
    assert ListValue.from_dict(list_value.to_dict()) == list_value


def test_nullvalue():
    from tests.outputs.google.google.protobuf import NullValue, Value

    null_value = NullValue.NULL_VALUE

    assert bytes(Value(null_value=null_value)) == b"\x08\x00"


def test_value_to_dict():
    from tests.outputs.google.google.protobuf import Value

    value = Value.from_dict([1, 2, False])

    assert value.to_dict() == [1, 2, False]
    assert Value.from_dict(value.to_dict()) == value



from datetime import (
    datetime,
    timedelta,
    timezone,
)

import pytest

from tests.outputs.timestamp_dict_encode.timestamp_dict_encode import Test

# Current World Timezone range (UTC-12 to UTC+14)
MIN_UTC_OFFSET_MIN = -12 * 60
MAX_UTC_OFFSET_MIN = 14 * 60

# Generate all timezones in range in 15 min increments
timezones = [timezone(timedelta(minutes=x)) for x in range(MIN_UTC_OFFSET_MIN, MAX_UTC_OFFSET_MIN + 1, 15)]


@pytest.mark.parametrize("tz", timezones)
def test_datetime_dict_encode(tz: timezone):
    original_time = datetime.now(tz=tz)
    original_message = Test()
    original_message.ts = original_time
    encoded = original_message.to_dict()
    decoded_message = Test.from_dict(encoded)

    # check that the timestamps are equal after decoding from dict
    assert original_message.ts.tzinfo is not None
    assert decoded_message.ts.tzinfo is not None
    assert original_message.ts == decoded_message.ts


@pytest.mark.parametrize("tz", timezones)
def test_json_serialize(tz: timezone):
    original_time = datetime.now(tz=tz)
    original_message = Test()
    original_message.ts = original_time
    json_serialized = original_message.to_json()
    decoded_message = Test.from_json(json_serialized)

    # check that the timestamps are equal after decoding from dict
    assert original_message.ts.tzinfo is not None
    assert decoded_message.ts.tzinfo is not None
    assert original_message.ts == decoded_message.ts



import copy
import importlib
import json
import math
import os
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import pytest

import betterproto2
from tests.util import requires_grpcio, requires_grpclib, requires_protobuf, requires_pydantic  # noqa: F401

# Force pure-python implementation instead of C++, otherwise imports
# break things because we can't properly reset the symbol database.
os.environ["PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION"] = "python"


@dataclass
class TestCase:
    jsons: list[str]
    plugin_package: str
    reference_package: str
    reference_path: list[str] | None = None
    xfail: bool = False


def list_replace_nans(items: list) -> list[Any]:
    """Replace float("nan") in a list with the string "NaN"

    Parameters
    ----------
    items : List
            List to update

    Returns
    -------
    List[Any]
        Updated list
    """
    result = []
    for item in items:
        if isinstance(item, list):
            result.append(list_replace_nans(item))
        elif isinstance(item, dict):
            result.append(dict_replace_nans(item))
        elif isinstance(item, float) and math.isnan(item):
            result.append(betterproto2.NAN)
    return result


def dict_replace_nans(input_dict: dict[Any, Any]) -> dict[Any, Any]:
    """Replace float("nan") in a dictionary with the string "NaN"

    Parameters
    ----------
    input_dict : Dict[Any, Any]
            Dictionary to update

    Returns
    -------
    Dict[Any, Any]
        Updated dictionary
    """
    result = {}
    for key, value in input_dict.items():
        if isinstance(value, dict):
            value = dict_replace_nans(value)
        elif isinstance(value, list):
            value = list_replace_nans(value)
        elif isinstance(value, float) and math.isnan(value):
            value = betterproto2.NAN
        result[key] = value
    return result


@pytest.fixture
def reset_sys_path():
    original = copy.deepcopy(sys.path)
    yield
    sys.path = original


TEST_CASES = [
    TestCase(["bool/bool.json"], "bool.bool", "bool_reference.bool_pb2"),
    TestCase(["bytes/bytes.json"], "bytes.bytes", "bytes_reference.bytes_pb2"),
    TestCase(["casing/casing.json"], "casing.casing", "casing_reference.casing_pb2"),
    TestCase(["deprecated/deprecated.json"], "deprecated.deprecated", "deprecated_reference.deprecated_pb2"),
    TestCase(["double/double.json", "double/double-negative.json"], "double.double", "double_reference.double_pb2"),
    TestCase(["enum/enum.json"], "enum.enum", "enum_reference.enum_pb2"),
    TestCase(
        ["field_name_identical_to_type/field_name_identical_to_type.json"],
        "field_name_identical_to_type.field_name_identical_to_type",
        "field_name_identical_to_type_reference.field_name_identical_to_type_pb2",
    ),
    TestCase(["fixed/fixed.json"], "fixed.fixed", "fixed_reference.fixed_pb2"),
    TestCase(["float/float.json"], "float.float", "float_reference.float_pb2"),
    TestCase(
        ["googletypes/googletypes.json", "googletypes/googletypes-missing.json"],
        "googletypes.googletypes",
        "googletypes_reference.googletypes_pb2",
    ),
    TestCase(
        ["googletypes_struct/googletypes_struct.json"],
        "googletypes_struct.googletypes_struct",
        "googletypes_struct_reference.googletypes_struct_pb2",
    ),
    TestCase(
        ["googletypes_value/googletypes_value.json"],
        "googletypes_value.googletypes_value",
        "googletypes_value_reference.googletypes_value_pb2",
    ),
    TestCase(["int32/int32.json"], "int32.int32", "int32_reference.int32_pb2"),
    TestCase(["map/map.json"], "map.map", "map_reference.map_pb2"),
    TestCase(["mapmessage/mapmessage.json"], "mapmessage.mapmessage", "mapmessage_reference.mapmessage_pb2"),
    TestCase(
        ["namespace_builtin_types/namespace_builtin_types.json"],
        "namespace_builtin_types.namespace_builtin_types",
        "namespace_builtin_types_reference.namespace_builtin_types_pb2",
    ),
    TestCase(
        ["namespace_keywords/namespace_keywords.json"],
        "namespace_keywords.namespace_keywords",
        "namespace_keywords_reference.namespace_keywords_pb2",
        xfail=True,
    ),
    TestCase(["nested/nested.json"], "nested.nested", "nested_reference.nested_pb2"),
    TestCase(["nestedtwice/nestedtwice.json"], "nestedtwice.nestedtwice", "nestedtwice_reference.nestedtwice_pb2"),
    TestCase(
        ["oneof_empty/oneof_empty.json", "oneof_empty/oneof_empty_maybe1.json", "oneof_empty/oneof_empty_maybe2.json"],
        "oneof_empty.oneof_empty",
        "oneof_empty_reference.oneof_empty_pb2",
    ),
    TestCase(
        ["oneof_enum/oneof_enum-enum-0.json", "oneof_enum/oneof_enum-enum-1.json", "oneof_enum/oneof_enum.json"],
        "oneof_enum.oneof_enum",
        "oneof_enum_reference.oneof_enum_pb2",
    ),
    TestCase(
        ["oneof/oneof.json", "oneof/oneof-name.json", "oneof/oneof_name.json"],
        "oneof.oneof",
        "oneof_reference.oneof_pb2",
    ),
    TestCase(
        ["proto3_field_presence_oneof/proto3_field_presence_oneof.json"],
        "proto3_field_presence_oneof.proto3_field_presence_oneof",
        "proto3_field_presence_oneof_reference.proto3_field_presence_oneof_pb2",
    ),
    TestCase(
        [
            "proto3_field_presence/proto3_field_presence_default.json",
            "proto3_field_presence/proto3_field_presence.json",
            "proto3_field_presence/proto3_field_presence_missing.json",
        ],
        "proto3_field_presence.proto3_field_presence",
        "proto3_field_presence_reference.proto3_field_presence_pb2",
    ),
    TestCase(
        ["recursivemessage/recursivemessage.json"],
        "recursivemessage.recursivemessage",
        "recursivemessage_reference.recursivemessage_pb2",
    ),
    TestCase(["ref/ref.json"], "ref.ref", "ref_reference.ref_pb2", reference_path=["ref_reference"]),
    TestCase(
        ["repeated_duration_timestamp/repeated_duration_timestamp.json"],
        "repeated_duration_timestamp.repeated_duration_timestamp",
        "repeated_duration_timestamp_reference.repeated_duration_timestamp_pb2",
    ),
    TestCase(
        ["repeatedmessage/repeatedmessage.json"],
        "repeatedmessage.repeatedmessage",
        "repeatedmessage_reference.repeatedmessage_pb2",
    ),
    TestCase(
        ["repeatedpacked/repeatedpacked.json"],
        "repeatedpacked.repeatedpacked",
        "repeatedpacked_reference.repeatedpacked_pb2",
    ),
    TestCase(["repeated/repeated.json"], "repeated.repeated", "repeated_reference.repeated_pb2"),
    TestCase(["signed/signed.json"], "signed.signed", "signed_reference.signed_pb2"),
    TestCase(
        ["timestamp_dict_encode/timestamp_dict_encode.json"],
        "timestamp_dict_encode.timestamp_dict_encode",
        "timestamp_dict_encode_reference.timestamp_dict_encode_pb2",
    ),
]


@pytest.mark.parametrize("test_case", TEST_CASES, ids=lambda x: x.plugin_package)
def test_message_json(test_case: TestCase, requires_pydantic, requires_grpcio, requires_grpclib) -> None:
    if test_case.xfail:
        pytest.xfail(f"Test case {test_case.plugin_package} is expected to fail.")

    plugin_module = importlib.import_module(f"tests.outputs.{test_case.plugin_package}")

    current_dir = Path(os.path.dirname(os.path.abspath(__file__)))

    for json_path in test_case.jsons:
        with open(current_dir / "inputs" / json_path) as f:
            json_data = f.read()

        message: betterproto2.Message = plugin_module.Test.from_json(json_data)
        message_json = message.to_json(indent=0)

        assert dict_replace_nans(json.loads(message_json)) == dict_replace_nans(json.loads(json_data))


@pytest.mark.parametrize("test_case", TEST_CASES, ids=lambda x: x.plugin_package)
def test_binary_compatibility(
    test_case: TestCase, reset_sys_path, requires_grpcio, requires_protobuf, requires_grpclib
) -> None:
    from google.protobuf.json_format import Parse

    if test_case.xfail:
        pytest.xfail(f"Test case {test_case.plugin_package} is expected to fail.")

    current_dir = Path(os.path.dirname(os.path.abspath(__file__)))

    if test_case.reference_path:
        for path in test_case.reference_path:
            sys.path.append(str(current_dir / "outputs" / path))

    plugin_module = importlib.import_module(f"tests.outputs.{test_case.plugin_package}")
    reference_module = importlib.import_module(f"tests.outputs.{test_case.reference_package}")

    # TODO fix and delete
    if "map" in plugin_module.__file__.replace("\\", "/").split("/"):
        pytest.skip("Skipping this test for now.")

    for json_path in test_case.jsons:
        with open(current_dir / "inputs" / json_path) as f:
            json_data = f.read()

        reference_instance = Parse(json_data, reference_module.Test())
        reference_binary_output = reference_instance.SerializeToString()

        plugin_instance_from_json: betterproto2.Message = plugin_module.Test().from_json(json_data)
        plugin_instance_from_binary = plugin_module.Test.FromString(reference_binary_output)

        # Generally this can't be relied on, but here we are aiming to match the
        # existing Python implementation and aren't doing anything tricky.
        # https://developers.google.com/protocol-buffers/docs/encoding#implications
        assert bytes(plugin_instance_from_json) == reference_binary_output
        assert bytes(plugin_instance_from_binary) == reference_binary_output

        assert plugin_instance_from_json == plugin_instance_from_binary
        assert dict_replace_nans(plugin_instance_from_json.to_dict()) == dict_replace_nans(
            plugin_instance_from_binary.to_dict()
        )



import pytest

from tests.util import requires_pydantic  # noqa: F401


def test_value():
    from tests.outputs.bool.bool import Test

    message = Test()
    assert not message.value, "Boolean is False by default"


def test_pydantic_no_value(requires_pydantic):
    from tests.outputs.bool_pydantic.bool import Test as TestPyd

    message = TestPyd()
    assert not message.value, "Boolean is False by default"


def test_pydantic_value(requires_pydantic):
    from tests.outputs.bool_pydantic.bool import Test as TestPyd

    message = TestPyd(value=False)
    assert not message.value


def test_pydantic_bad_value(requires_pydantic):
    from tests.outputs.bool_pydantic.bool import Test as TestPyd

    with pytest.raises(ValueError):
        TestPyd(value=123)



import sys

import pytest


@pytest.mark.skipif(
    sys.version_info < (3, 10),
    reason="pattern matching is only supported in python3.10+",
)
def test_oneof_pattern_matching():
    from tests.oneof_pattern_matching import test_oneof_pattern_matching

    test_oneof_pattern_matching()



import asyncio
import threading
from collections.abc import AsyncIterator

import pytest

from tests.util import requires_grpcio, requires_grpclib  # noqa: F401


@pytest.mark.asyncio
async def test_sync_client(requires_grpcio, requires_grpclib):
    import grpc
    from grpclib.server import Server

    from tests.outputs.simple_service.simple_service import Request, Response, SimpleServiceBase, SimpleServiceSyncStub

    class SimpleService(SimpleServiceBase):
        async def get_unary_unary(self, message: "Request") -> "Response":
            return Response(message=f"Hello {message.value}")

        async def get_unary_stream(self, message: "Request") -> "AsyncIterator[Response]":
            for i in range(5):
                yield Response(message=f"Hello {message.value} {i}")

        async def get_stream_unary(self, messages: "AsyncIterator[Request]") -> "Response":
            s = 0
            async for m in messages:
                s += m.value
            return Response(message=f"Hello {s}")

        async def get_stream_stream(self, messages: "AsyncIterator[Request]") -> "AsyncIterator[Response]":
            async for message in messages:
                yield Response(message=f"Hello {message.value}")

    start_server_event = threading.Event()
    close_server_event = asyncio.Event()

    def start_server():
        async def run_server():
            server = Server([SimpleService()])
            await server.start("127.0.0.1", 1234)
            start_server_event.set()

            await close_server_event.wait()
            server.close()

        loop = asyncio.new_event_loop()
        loop.run_until_complete(run_server())
        loop.close()

    # We need to start the server in a new thread to avoid a deadlock
    server_thread = threading.Thread(target=start_server)
    server_thread.start()

    # Create a sync client
    start_server_event.wait()

    with grpc.insecure_channel("localhost:1234") as channel:
        client = SimpleServiceSyncStub(channel)

        response = client.get_unary_unary(Request(value=42))
        assert response.message == "Hello 42"

        response = client.get_unary_stream(Request(value=42))
        assert [r.message for r in response] == [f"Hello 42 {i}" for i in range(5)]

        response = client.get_stream_unary([Request(value=i) for i in range(5)])
        assert response.message == "Hello 10"

        response = client.get_stream_stream([Request(value=i) for i in range(5)])
        assert [r.message for r in response] == [f"Hello {i}" for i in range(5)]

        close_server_event.set()

    # Create an async client
    # client = SimpleServiceStub(Channel(host="127.0.0.1", port=1234))
    # response = await client.get_unary_unary(Request(value=42))
    # assert response.message == "Hello 42"



import pytest

from tests.util import requires_grpclib  # noqa: F401


@pytest.mark.asyncio
async def test_rpc_input_message(requires_grpclib):
    from grpclib.testing import ChannelFor

    from tests.outputs.rpc_empty_input_message.rpc_empty_input_message import (
        Response,
        ServiceBase,
        ServiceStub,
        Test,
    )

    class Service(ServiceBase):
        async def read(self, test: "Test") -> "Response":
            return Response(v=42)

    async with ChannelFor([Service()]) as channel:
        client = ServiceStub(channel)

        assert (await client.read(Test())).v == 42

        # Check that we can call the method without providing the message
        assert (await client.read()).v == 42



import pytest

from tests.util import requires_pydantic  # noqa: F401


def test_manual_validation(requires_pydantic):
    import pydantic

    from tests.outputs.manual_validation_pydantic.manual_validation import Msg

    msg = Msg()

    msg.x = 12
    msg._validate()

    msg.x = 2**50  # This is an invalid int32 value
    with pytest.raises(pydantic.ValidationError):
        msg._validate()


def test_manual_validation_non_pydantic():
    from tests.outputs.manual_validation.manual_validation import Msg

    # Validation is not available for non-pydantic messages
    with pytest.raises(TypeError):
        Msg()._validate()



import inspect

from tests.util import requires_grpclib  # noqa: F401


def test_parameters(requires_grpclib):
    from tests.outputs.service_uppercase.service_uppercase import TestStub

    sig = inspect.signature(TestStub.do_thing)
    assert len(sig.parameters) == 5, "Expected 5 parameters"



import pytest

import betterproto2


class Colour(betterproto2.Enum):
    RED = 1
    GREEN = 2
    BLUE = 3


PURPLE = Colour(4)


@pytest.mark.parametrize(
    "member, str_value",
    [
        (Colour.RED, "RED"),
        (Colour.GREEN, "GREEN"),
        (Colour.BLUE, "BLUE"),
        (PURPLE, "UNKNOWN(4)"),
    ],
)
def test_str(member: Colour, str_value: str) -> None:
    assert str(member) == str_value


@pytest.mark.parametrize(
    "member, repr_value",
    [
        (Colour.RED, "<Colour.RED: 1>"),
        (Colour.GREEN, "<Colour.GREEN: 2>"),
        (Colour.BLUE, "<Colour.BLUE: 3>"),
        (PURPLE, "<Colour.~UNKNOWN: 4>"),
    ],
)
def test_repr(member: Colour, repr_value: str) -> None:
    assert repr(member) == repr_value


@pytest.mark.parametrize(
    "member, values",
    [
        (Colour.RED, ("RED", 1)),
        (Colour.GREEN, ("GREEN", 2)),
        (Colour.BLUE, ("BLUE", 3)),
        (PURPLE, ("", 4)),
    ],
)
def test_name_values(member: Colour, values: tuple[str | None, int]) -> None:
    assert (member.name, member.value) == values


@pytest.mark.parametrize(
    "member, input_str",
    [
        (Colour.RED, "RED"),
        (Colour.GREEN, "GREEN"),
        (Colour.BLUE, "BLUE"),
    ],
)
def test_from_string(member: Colour, input_str: str) -> None:
    assert Colour.from_string(input_str) == member


@pytest.mark.parametrize(
    "member, input_int",
    [
        (Colour.RED, 1),
        (Colour.GREEN, 2),
        (Colour.BLUE, 3),
        (PURPLE, 4),
    ],
)
def test_construction(member: Colour, input_int: int) -> None:
    assert Colour(input_int) == member


def test_enum_renaming() -> None:
    from tests.outputs.enum.enum import ArithmeticOperator, HttpCode, NoStriping

    assert set(ArithmeticOperator.__members__) == {"NONE", "PLUS", "MINUS", "_0_PREFIXED"}
    assert set(HttpCode.__members__) == {"UNSPECIFIED", "OK", "NOT_FOUND"}
    assert set(NoStriping.__members__) == {"NO_STRIPING_NONE", "NO_STRIPING_A", "B"}


def test_enum_to_dict() -> None:
    from tests.outputs.enum.enum import ArithmeticOperator, EnumMessage, NoStriping

    msg = EnumMessage(
        arithmetic_operator=ArithmeticOperator.PLUS,
        no_striping=NoStriping.NO_STRIPING_A,
    )

    assert msg.to_dict() == {
        "arithmeticOperator": "ARITHMETIC_OPERATOR_PLUS",  # The original proto name must be preserved
        "noStriping": "NO_STRIPING_A",
    }

    assert EnumMessage.from_dict(msg.to_dict()) == msg


def test_unknown_variant_to_dict() -> None:
    from tests.outputs.enum.enum import NewVersion, NewVersionMessage, OldVersionMessage

    serialized = bytes(NewVersionMessage(new_version=NewVersion.V3))

    deserialized = OldVersionMessage.parse(serialized)

    assert deserialized.to_dict() == {"oldVersion": 3}



def test_nested_from_dict():
    """
    Make sure that from_dict() arguments are passed recursively
    """
    from tests.outputs.nested.nested import Test

    data = {
        "nested": {"count": 1},
        "sibling": {"foo": 2},
    }
    Test.from_dict(data)

    data["bar"] = 3
    Test.from_dict(data, ignore_unknown_fields=True)

    data["nested"]["bar"] = 3
    Test.from_dict(data, ignore_unknown_fields=True)

    data["sibling"]["bar"] = 4
    Test.from_dict(data, ignore_unknown_fields=True)



import pytest

from tests.util import requires_pydantic  # noqa: F401


def test_invalid_field():
    from tests.outputs.invalid_field.invalid_field import Test

    with pytest.raises(TypeError):
        Test(unknown_field=12)


def test_invalid_field_pydantic(requires_pydantic):
    from pydantic import ValidationError

    from tests.outputs.invalid_field_pydantic.invalid_field import Test

    with pytest.raises(ValidationError):
        Test(unknown_field=12)



from collections.abc import Callable
from typing import TYPE_CHECKING, Any

import pytest

from tests.util import requires_grpclib  # noqa: F401

if TYPE_CHECKING:
    from tests.outputs.googletypes_response.googletypes_response import Input, TestStub


def get_test_cases() -> list[tuple[Callable[["TestStub", "Input"], Any], Callable, Any]]:
    import tests.outputs.googletypes_response.google.protobuf as protobuf
    from tests.outputs.googletypes_response.googletypes_response import TestStub

    return [
        (TestStub.get_double, protobuf.DoubleValue, 2.5),
        (TestStub.get_float, protobuf.FloatValue, 2.5),
        (TestStub.get_int_64, protobuf.Int64Value, -64),
        (TestStub.get_u_int_64, protobuf.UInt64Value, 64),
        (TestStub.get_int_32, protobuf.Int32Value, -32),
        (TestStub.get_u_int_32, protobuf.UInt32Value, 32),
        (TestStub.get_bool, protobuf.BoolValue, True),
        (TestStub.get_string, protobuf.StringValue, "string"),
        (TestStub.get_bytes, protobuf.BytesValue, bytes(0xFF)[0:4]),
    ]


@pytest.mark.asyncio
async def test_channel_receives_wrapped_type(requires_grpclib):
    from tests.mocks import MockChannel
    from tests.outputs.googletypes_response.googletypes_response import Input, TestStub

    for service_method, wrapper_class, value in get_test_cases():
        wrapped_value = wrapper_class()
        wrapped_value.value = value
        channel = MockChannel(responses=[wrapped_value])
        service = TestStub(channel)
        method_param = Input()

        await service_method(service, method_param)

        assert channel.requests[0]["response_type"] != type(value) | None
        assert channel.requests[0]["response_type"] == type(wrapped_value)


@pytest.mark.asyncio
@pytest.mark.xfail
async def test_service_unwraps_response(requires_grpclib):
    """
    grpclib does not unwrap wrapper values returned by services
    """
    from tests.mocks import MockChannel
    from tests.outputs.googletypes_response.googletypes_response import Input, TestStub

    for service_method, wrapper_class, value in get_test_cases():
        wrapped_value = wrapper_class()
        wrapped_value.value = value
        service = TestStub(MockChannel(responses=[wrapped_value]))
        method_param = Input()

        response_value = await service_method(service, method_param)

        assert response_value == value
        assert type(response_value) == type(value)
